@@ -1,7 +1,6 @@
 #include <linux/version.h>
-#ifdef HAVE_EXPORT_H
-#include <linux/export.h>
-#endif
+#define IP_SET_COMPAT_HEADERS
+#include <linux/netfilter/ipset/ip_set_compat.h>
 #include <linux/netfilter/ipset/pfxlen.h>
 
 /* Prefixlen maps for fast conversions, by Jan Engelhardt. */
