@@ -1,6 +1,6 @@
 #!/bin/sh
 
-set -x
+# set -x
 set -e
 
 ipset=${IPSET_BIN:-../src/ipset}
@@ -111,6 +111,12 @@ mangle)
 	$cmd -t mangle -A INPUT -m mark --mark 0x1234 -j LOG --log-prefix "in set mark: "
 	$cmd -t mangle -A INPUT -s 10.255.0.0/16 -j DROP
 	;;
+netiface)
+	$ipset n test hash:net,iface
+	$ipset a test 0.0.0.0/0,eth0
+	$cmd -A OUTPUT -m set --match-set test dst,dst -j LOG --log-prefix "in set netiface: "
+	$cmd -A OUTPUT -d 10.255.255.254 -j DROP
+	;;
 stop)
 	$cmd -F
 	$cmd -X
